@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2004 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2004 Laszlo Molnar
+   Copyright (C) 1996-2006 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2006 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -71,12 +71,11 @@ void init_options(struct options_t *o)
     for (unsigned i = 0; i < TABLESIZE(o->win32_pe.compress_rt); i++)
         o->win32_pe.compress_rt[i] = -1;
     o->win32_pe.compress_rt[24] = false;    // 24 == RT_MANIFEST
-    o->win32_pe.strip_loadconf = -1;
     o->win32_pe.strip_relocs = -1;
 }
 
 static struct options_t global_options;
-struct options_t * volatile opt = &global_options;
+struct options_t *opt = &global_options;
 
 static int done_output_name = 0;
 static int done_script_name = 0;
@@ -444,7 +443,7 @@ static int do_option(int optc, const char *arg)
         set_cmd(CMD_DECOMPRESS);
         break;
     case 'D':
-        opt->debug++;
+        opt->debug.debug_level++;
         break;
     case 'f':
         opt->force++;
@@ -538,6 +537,23 @@ static int do_option(int optc, const char *arg)
             e_method(opt->method, 10);
         break;
 
+    // debug
+    case 542:
+        if (!mfx_optarg || strlen(mfx_optarg) != 4)
+            e_optarg(arg);
+        memcpy(opt->debug.fake_stub_version, mfx_optarg, 4);
+        break;
+    case 543:
+        if (!mfx_optarg || strlen(mfx_optarg) != 4)
+            e_optarg(arg);
+        memcpy(opt->debug.fake_stub_year, mfx_optarg, 4);
+        break;
+    case 544:
+        if (!mfx_optarg || !mfx_optarg[0])
+            e_optarg(arg);
+        opt->debug.dump_stub_loader = mfx_optarg;
+        break;
+
     // misc
     case 512:
         opt->console = CON_FILE;
@@ -678,10 +694,7 @@ static int do_option(int optc, const char *arg)
         //printf("compress_resources: %d\n", opt->win32_pe.compress_resources);
         break;
     case 633:
-        opt->win32_pe.strip_loadconf = 1;
-        if (mfx_optarg && strcmp(mfx_optarg,"0") == 0)
-            opt->win32_pe.strip_loadconf = 0;
-        //printf("strip_loadconf: %d\n", opt->win32_pe.strip_loadconf);
+        // opt->win32_pe.strip_loadconf - OBSOLETE - IGNORED
         break;
     case 634:
         opt->win32_pe.strip_relocs = 1;
@@ -757,7 +770,6 @@ static const struct mfx_option longopts[] =
     {"version",             0, 0, 'V'+256}, // display version number
 
     // options
-    {"debug",            0x10, 0, 'D'},
     {"force",               0, 0, 'f'},     // force overwrite of output files
     {"force-compress",      0, 0, 'f'},     //   and compression of suspicious files
     {"info",                0, 0, 'i'},     // info mode
@@ -773,6 +785,12 @@ static const struct mfx_option longopts[] =
 #endif
     {"verbose",             0, 0, 'v'},     // verbose mode
 
+    // debug options
+    {"debug",            0x10, 0, 'D'},
+    {"dump-stub-loader" ,0x31, 0, 544},     // for internal debugging
+    {"fake-stub-version",0x31, 0, 542},     // for internal debugging
+    {"fake-stub-year"   ,0x31, 0, 543},     // for internal debugging
+
     // backup options
     {"backup",              0, 0, 'k'},
     {"keep",                0, 0, 'k'},
@@ -839,7 +857,7 @@ static const struct mfx_option longopts[] =
     {"compress-exports",    2, 0, 630},
     {"compress-icons",      2, 0, 631},
     {"compress-resources",  2, 0, 632},
-    {"strip-loadconf",   0x12, 0, 633},
+    {"strip-loadconf",   0x12, 0, 633},     // OBSOLETE - IGNORED
     {"strip-relocs",     0x12, 0, 634},
     // ps1/exe
     {"boot-only",        0x10, 0, 670},
@@ -921,7 +939,7 @@ static const struct mfx_option longopts[] =
     {"compress-exports",    2, 0, 630},
     {"compress-icons",      2, 0, 631},
     {"compress-resources",  2, 0, 632},
-    {"strip-loadconf",   0x12, 0, 633},
+    {"strip-loadconf",   0x12, 0, 633},     // OBSOLETE - IGNORED
     {"strip-relocs",     0x12, 0, 634},
 
     { NULL, 0, NULL, 0 }
@@ -1094,6 +1112,10 @@ void upx_sanity_check(void)
     COMPILE_TIME_ASSERT(sizeof(UPX_VERSION_STRING4) == 4 + 1)
     assert(strlen(UPX_VERSION_STRING4) == 4);
     assert(memcmp(UPX_VERSION_STRING4, UPX_VERSION_STRING, 4) == 0);
+    COMPILE_TIME_ASSERT(sizeof(UPX_VERSION_YEAR) == 4 + 1)
+    assert(strlen(UPX_VERSION_YEAR) == 4);
+    assert(memcmp(UPX_VERSION_DATE_ISO, UPX_VERSION_YEAR, 4) == 0);
+    assert(memcmp(UPX_VERSION_DATE + strlen(UPX_VERSION_DATE) - 4, UPX_VERSION_YEAR, 4) == 0);
 
 #if 1
 #  if 1
