@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2004 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2004 Laszlo Molnar
+   Copyright (C) 1996-2006 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2006 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -158,6 +158,7 @@ PackArmPe::PackArmPe(InputFile *f) : super(f)
     kernel32ordinal = false;
     tlsindex = 0;
     big_relocs = 0;
+    sorelocs = 0;
     soxrelocs = 0;
 }
 
@@ -565,7 +566,6 @@ unsigned PackArmPe::processImports() // pass 1
         LE32       *lookupt;
         unsigned   npos;
         bool       isk32;
-        unsigned   _;           // padding to 32
 
         static int __acc_cdecl_qsort compare(const void *p1, const void *p2)
         {
@@ -581,8 +581,7 @@ unsigned PackArmPe::processImports() // pass 1
             if (!u2->shname) return -1;
             return strlen(u1->shname) - strlen(u2->shname);
         }
-    }
-    __attribute_packed;
+    };
 
     // +1 for dllnum=0
     Array(struct udll, dlls, dllnum+1);
@@ -611,7 +610,6 @@ unsigned PackArmPe::processImports() // pass 1
                 importbyordinal = true;
                 soimport += 2; // ordinal num: 2 bytes
                 dlls[ic].ordinal = *tarr & 0xffff;
-
                 //if (dlls[ic].isk32)
                 //    kernel32ordinal = true,k32o++;
             }
@@ -658,8 +656,8 @@ unsigned PackArmPe::processImports() // pass 1
     strcpy(dllnames,kernel32dll);
     im->dllname = k32namepos;
     im->iat = ptr_diff(ordinals,oimpdlls);
-    *ordinals++ = ptr_diff(importednames,oimpdlls);
-    *ordinals++ = ptr_diff(importednames,oimpdlls) + 14;
+    *ordinals++ = ptr_diff(importednames,oimpdlls);             // LoadLibraryW
+    *ordinals++ = ptr_diff(importednames,oimpdlls) + 14;        // GetProcAddressA
     dllnames += sizeof(kernel32dll);
     importednames += sizeof(llgpa);
 
@@ -672,7 +670,7 @@ unsigned PackArmPe::processImports() // pass 1
             if (idlls[ic]->ordinal)
                 for (LE32 *tarr = idlls[ic]->lookupt; *tarr; tarr++)
                     if (*tarr & 0x80000000)
-                    *ordinals++ = *tarr;
+                        *ordinals++ = *tarr;
             */
         }
         else if (ic && strcasecmp(idlls[ic-1]->name,idlls[ic]->name) == 0)
@@ -702,6 +700,7 @@ unsigned PackArmPe::processImports() // pass 1
 
     Interval names(ibuf),iats(ibuf),lookups(ibuf);
     // create the preprocessed data
+    //ordinals -= k32o;
     upx_byte *ppi = oimport;  // preprocessed imports
     for (ic = 0; ic < dllnum; ic++)
     {
@@ -1550,9 +1549,6 @@ int PackArmPe::buildLoader(const Filter *ft)
 
 void PackArmPe::pack(OutputFile *fo)
 {
-    if (opt->win32_pe.strip_loadconf < 0)
-        opt->win32_pe.strip_loadconf = false;
-
     const unsigned objs = ih.objects;
     isection = new pe_section_t[objs];
     fi->seek(pe_offset+sizeof(ih),SEEK_SET);
@@ -1565,14 +1561,15 @@ void PackArmPe::pack(OutputFile *fo)
     // check the PE header
     // FIXME: add more checks
     if (!opt->force && (
-        ih.opthdrsize != 0xE0
-        || (ih.flags & EXECUTABLE) == 0
-        || ih.subsystem != 9
+           (ih.cpu != 0x1c0)
+        || (ih.opthdrsize != 0xe0)
+        || ((ih.flags & EXECUTABLE) == 0)
+        || (ih.subsystem != 9)
         || (ih.entry == 0 && !isdll)
-        || ih.ddirsentries != 16
-      /*|| IDSIZE(PEDIR_EXCEPTION*/) // is this used on i386?
+        || (ih.ddirsentries != 16)
+//        || IDSIZE(PEDIR_EXCEPTION) // is this used on arm?
 //        || IDSIZE(PEDIR_COPYRIGHT)
-       )
+       ))
         throwCantPack("unexpected value in PE header (try --force)");
 
     if (IDSIZE(PEDIR_SEC))
@@ -1589,10 +1586,6 @@ void PackArmPe::pack(OutputFile *fo)
     if (IDSIZE(PEDIR_RELOC))
         throwCantPack("relocations not yet supported");
 
-    // Structured Exception Handling
-    if (!opt->win32_pe.strip_loadconf && IDSIZE(PEDIR_LOADCONF))
-        throwCantPack("Structured Exception Handling present (try --strip-loadconf)");
-
     if (isdll)
         opt->win32_pe.strip_relocs = false;
     else if (opt->win32_pe.strip_relocs < 0)
@@ -1882,12 +1875,6 @@ void PackArmPe::pack(OutputFile *fo)
     ODADDR(PEDIR_BOUNDIM) = 0;
     ODSIZE(PEDIR_BOUNDIM) = 0;
 
-    if (opt->win32_pe.strip_loadconf)
-    {
-        ODADDR(PEDIR_LOADCONF) = 0;
-        ODSIZE(PEDIR_LOADCONF) = 0;
-    }
-
     // tls is put into section 1
 
     ic = s1addr + s1size - sotls;
@@ -1904,10 +1891,12 @@ void PackArmPe::pack(OutputFile *fo)
     ODADDR(PEDIR_RESOURCE) = soresources ? ic : 0;
     ODSIZE(PEDIR_RESOURCE) = soresources;
     ic += soresources;
+
     processImports(ic);
     ODADDR(PEDIR_IMPORT) = ic;
     ODSIZE(PEDIR_IMPORT) = soimpdlls;
     ic += soimpdlls;
+
     //processExports(&xport,ic);
     ODADDR(PEDIR_EXPORT) = soexport ? ic : 0;
     ODSIZE(PEDIR_EXPORT) = soexport;
@@ -1917,6 +1906,7 @@ void PackArmPe::pack(OutputFile *fo)
         ODSIZE(PEDIR_EXPORT) = IDSIZE(PEDIR_EXPORT);
     }
     ic += soexport;
+
     //processRelocs(&rel);
     ODADDR(PEDIR_RELOC) = soxrelocs ? ic : 0;
     ODSIZE(PEDIR_RELOC) = soxrelocs;
@@ -1999,6 +1989,8 @@ void PackArmPe::pack(OutputFile *fo)
     fo->write(obuf,clen);
     infoWriting("compressed data", clen);
     fo->write(loader,codesize);
+    if (opt->debug.dump_stub_loader)
+        OutputFile::dump(opt->debug.dump_stub_loader, loader, codesize);
     if ((ic = fo->getBytesWritten() & 3) != 0)
         fo->write(ibuf,4 - ic);
     fo->write(otls,sotls);
@@ -2047,8 +2039,11 @@ int PackArmPe::canUnpack()
 }
 
 void PackArmPe::unpack(OutputFile *)
-{}
+{
+}
+
 
 /*
 vi:ts=4:et
 */
+
