@@ -304,9 +304,19 @@ void PackUnix::packExtent(
     unsigned &total_in,
     unsigned &total_out,
     Filter *ft,
-    OutputFile *fo
+    OutputFile *fo,
+    unsigned hdr_ulen
 )
 {
+    unsigned const init_u_adler = ph.u_adler;
+    unsigned const init_c_adler = ph.c_adler;
+    MemBuffer hdr_ibuf;
+    if (hdr_ulen) {
+        hdr_ibuf.alloc(hdr_ulen);
+        fi->seek(0, SEEK_SET);
+        int l = fi->readx(hdr_ibuf, hdr_ulen);
+        (void)l;
+    }
     fi->seek(x.offset, SEEK_SET);
     for (off_t rest = x.size; 0 != rest; ) {
         int const strategy = getStrategy(*ft);
@@ -336,7 +346,9 @@ void PackUnix::packExtent(
             ft->id = 0;
             ft->cto = 0;
 
-            compressWithFilters(ft, OVERHEAD, strategy);
+            compressWithFilters(ft, OVERHEAD, strategy,
+                NULL, 0, 0, 0, 0, // those 5 args are the defaults
+                hdr_ibuf, hdr_ulen);
         }
         else {
             (void) compress(ibuf, obuf);    // ignore return value
@@ -352,12 +364,39 @@ void PackUnix::packExtent(
         if (ph.c_len >= ph.u_len) {
             // block is not compressible
             ph.c_len = ph.u_len;
+            memcpy(obuf, ibuf, ph.c_len);
             // must update checksum of compressed data
             ph.c_adler = upx_adler32(ibuf, ph.u_len, ph.saved_c_adler);
         }
 
         // write block sizes
         b_info tmp;
+        if (hdr_ulen) {
+            unsigned hdr_clen;
+            MemBuffer hdr_obuf;
+            unsigned result[16];
+            upx_compress_config_t conf;
+            memset(&conf, 0xff, sizeof(conf));
+            hdr_obuf.allocForCompression(hdr_ulen);
+            int r = upx_compress(hdr_ibuf, hdr_ulen, hdr_obuf, &hdr_clen, 0,
+                ph.method, 10, &conf, result);
+            (void)r;
+            ph.saved_u_adler = upx_adler32(hdr_ibuf, hdr_ulen, init_u_adler);
+            ph.saved_c_adler = upx_adler32(hdr_obuf, hdr_clen, init_c_adler);
+            ph.u_adler = upx_adler32(ibuf, ph.u_len, ph.saved_u_adler);
+            ph.c_adler = upx_adler32(obuf, ph.c_len, ph.saved_c_adler);
+            end_u_adler = ph.u_adler;
+            memset(&tmp, 0, sizeof(tmp));
+            set_native32(&tmp.sz_unc, hdr_ulen);
+            set_native32(&tmp.sz_cpr, hdr_clen);
+            tmp.b_method = (unsigned char) ph.method;
+            fo->write(&tmp, sizeof(tmp));
+            b_len += sizeof(b_info);
+            fo->write(hdr_obuf, hdr_clen);
+            total_out += hdr_clen;
+            total_in  += hdr_ulen;
+            hdr_ulen = 0;  // compress hdr one time only
+        }
         memset(&tmp, 0, sizeof(tmp));
         set_native32(&tmp.sz_unc, ph.u_len);
         set_native32(&tmp.sz_cpr, ph.c_len);
@@ -374,8 +413,8 @@ void PackUnix::packExtent(
         // write compressed data
         if (ph.c_len < ph.u_len) {
             fo->write(obuf, ph.c_len);
-            // Checks ph.u_adler after decompression but before unfiltering
-            verifyOverlappingDecompression();
+            // Checks ph.u_adler after decompression, after unfiltering
+            verifyOverlappingDecompression(ft);
         }
         else {
             fo->write(ibuf, ph.u_len);
