@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2004 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2004 Laszlo Molnar
+   Copyright (C) 1996-2006 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2006 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -159,7 +159,11 @@ PackW32Pe::PackW32Pe(InputFile *f) : super(f)
     kernel32ordinal = false;
     tlsindex = 0;
     big_relocs = 0;
+    sorelocs = 0;
+    soxrelocs = 0;
     soloadconf = 0;
+    use_dep_hack = true;
+    use_clear_dirty_stack = true;
 }
 
 
@@ -547,7 +551,9 @@ void PackW32Pe::processImports(unsigned myimport) // pass 2
 unsigned PackW32Pe::processImports() // pass 1
 {
     static const unsigned char kernel32dll[] = "KERNEL32.DLL";
-    static const char llgpa[] = "\x0\x0""LoadLibraryA\x0\x0""GetProcAddress\x0\x0";
+    static const char llgpa[] = "\x0\x0""LoadLibraryA\x0\x0"
+                                "GetProcAddress\x0\x0"
+                                "VirtualProtect\x0\x0";
     static const char exitp[] = "ExitProcess\x0\x0\x0";
 
     unsigned dllnum = 0;
@@ -569,7 +575,6 @@ unsigned PackW32Pe::processImports() // pass 1
         LE32       *lookupt;
         unsigned   npos;
         bool       isk32;
-        unsigned   _;           // padding to 32
 
         static int __acc_cdecl_qsort compare(const void *p1, const void *p2)
         {
@@ -585,8 +590,7 @@ unsigned PackW32Pe::processImports() // pass 1
             if (!u2->shname) return -1;
             return strlen(u1->shname) - strlen(u2->shname);
         }
-    }
-    __attribute_packed;
+    };
 
     // +1 for dllnum=0
     Array(struct udll, dlls, dllnum+1);
@@ -651,22 +655,23 @@ unsigned PackW32Pe::processImports() // pass 1
     im = (import_desc*) oimpdlls;
 
     LE32 *ordinals = (LE32*) (oimpdlls + (dllnum2 + 1) * sizeof(import_desc));
-    LE32 *lookuptable = ordinals + 3 + k32o + (isdll ? 0 : 1);
+    LE32 *lookuptable = ordinals + 4 + k32o + (isdll ? 0 : 1);
     upx_byte *dllnames = ((upx_byte*) lookuptable) + (dllnum2 - 1) * 8;
     upx_byte *importednames = dllnames + (dllnamelen &~ 1);
 
     unsigned k32namepos = ptr_diff(dllnames,oimpdlls);
 
-    memcpy(importednames,llgpa,sizeof(llgpa));
+    memcpy(importednames, llgpa, sizeof(llgpa));
     if (!isdll)
-        memcpy(importednames + sizeof(llgpa) - 1,exitp,sizeof(exitp));
+        memcpy(importednames + sizeof(llgpa) - 1, exitp, sizeof(exitp));
     strcpy(dllnames,kernel32dll);
     im->dllname = k32namepos;
     im->iat = ptr_diff(ordinals,oimpdlls);
-    *ordinals++ = ptr_diff(importednames,oimpdlls);
-    *ordinals++ = ptr_diff(importednames,oimpdlls) + 14;
+    *ordinals++ = ptr_diff(importednames,oimpdlls);             // LoadLibraryA
+    *ordinals++ = ptr_diff(importednames,oimpdlls) + 14;        // GetProcAddress
+    *ordinals++ = ptr_diff(importednames,oimpdlls) + 14 + 16;   // VirtualProtect
     if (!isdll)
-        *ordinals++ = ptr_diff(importednames,oimpdlls) + sizeof(llgpa) - 3;
+        *ordinals++ = ptr_diff(importednames,oimpdlls) + sizeof(llgpa) - 3; // ExitProcess
     dllnames += sizeof(kernel32dll);
     importednames += sizeof(llgpa) - 2 + (isdll ? 0 : sizeof(exitp) - 1);
 
@@ -761,7 +766,7 @@ unsigned PackW32Pe::processImports() // pass 1
         soimport = 0;
 
     //OutputFile::dump("x0.imp", oimport, soimport);
-    //OutputFile::dump("x1.imp", oimpdlls, soimpdlss);
+    //OutputFile::dump("x1.imp", oimpdlls, soimpdlls);
 
     unsigned ilen = 0;
     names.flatten();
@@ -1482,25 +1487,47 @@ void PackW32Pe::processResources(Resource *res)
 // Load Configuration handling
 **************************************************************************/
 
-void PackW32Pe::processLoadConf()
+void PackW32Pe::processLoadConf(Interval *iv) // pass 1
 {
     if (IDSIZE(PEDIR_LOADCONF) == 0)
         return;
 
-    const upx_byte * const loadconf = ibuf + IDADDR(PEDIR_LOADCONF);
+    const unsigned lcaddr = IDADDR(PEDIR_LOADCONF);
+    const upx_byte * const loadconf = ibuf + lcaddr;
     soloadconf = get_le32(loadconf);
     if (soloadconf == 0)
         return;
     if (soloadconf > 256)
         throwCantPack("size of Load Configuration directory unexpected");
 
-    if (IDSIZE(PEDIR_RELOC))
-        throwCantPack("LoadConf + relocations yet not supported");
+    // if there were relocation entries referring to the load config table
+    // then we need them for the copy of the table too
+    unsigned pos,type;
+    Reloc rel(ibuf + IDADDR(PEDIR_RELOC), IDSIZE(PEDIR_RELOC));
+    while (rel.next(pos, type))
+        if (pos >= lcaddr && pos < lcaddr + soloadconf)
+        {
+            iv->add(pos - lcaddr, type);
+            // printf("loadconf reloc detected: %x\n", pos);
+        }
 
     oloadconf = new upx_byte[soloadconf];
     memcpy(oloadconf, loadconf, soloadconf);
 }
 
+void PackW32Pe::processLoadConf(Reloc *rel, const Interval *iv,
+                                unsigned newaddr) // pass2
+{
+    // now we have the address of the new load config table
+    // so we can create the new relocation entries
+    for (unsigned ic = 0; ic < iv->ivnum; ic++)
+    {
+        rel->add(iv->ivarr[ic].start + newaddr, iv->ivarr[ic].len);
+        //printf("loadconf reloc added: %x %d\n",
+        //       iv->ivarr[ic].start + newaddr, iv->ivarr[ic].len);
+    }
+}
+
 //
 
 unsigned PackW32Pe::virta2objnum(unsigned addr,pe_section_t *sect,unsigned objs)
@@ -1618,8 +1645,13 @@ int PackW32Pe::buildLoader(const Filter *ft)
                   NULL
                  );
     }
-    addLoader("PEMAIN20",
-              ih.entry ? "PEDOJUMP" : "PERETURN",
+    if (use_dep_hack)
+        addLoader("PEDEPHAK", NULL);
+    addLoader("PEMAIN20", NULL);
+    if (use_clear_dirty_stack)
+        addLoader(ih.entry ? "PEDOJUMP_CLEARSTACK" : "PERETURN_CLEARSTACK", NULL);
+    addLoader("PEMAIN21", NULL);
+    addLoader(ih.entry ? "PEDOJUMP" : "PERETURN",
               "IDENTSTR,UPX1HEAD",
               NULL
              );
@@ -1629,9 +1661,6 @@ int PackW32Pe::buildLoader(const Filter *ft)
 
 void PackW32Pe::pack(OutputFile *fo)
 {
-    if (opt->win32_pe.strip_loadconf < 0)
-        opt->win32_pe.strip_loadconf = false;
-
     const unsigned objs = ih.objects;
     isection = new pe_section_t[objs];
     fi->seek(pe_offset+sizeof(ih),SEEK_SET);
@@ -1644,12 +1673,12 @@ void PackW32Pe::pack(OutputFile *fo)
     // check the PE header
     // FIXME: add more checks
     if (!opt->force && (
-           ih.cpu < 0x14c || ih.cpu > 0x150
-        || ih.opthdrsize != 0xE0
-        || (ih.flags & EXECUTABLE) == 0
+           (ih.cpu < 0x14c || ih.cpu > 0x150)
+        || (ih.opthdrsize != 0xe0)
+        || ((ih.flags & EXECUTABLE) == 0)
         || (ih.subsystem != 2 && ih.subsystem != 3 && ih.subsystem != 1)
         || (ih.entry == 0 && !isdll)
-        || ih.ddirsentries != 16
+        || (ih.ddirsentries != 16)
         || IDSIZE(PEDIR_EXCEPTION) // is this used on i386?
 //        || IDSIZE(PEDIR_COPYRIGHT)
        ))
@@ -1660,12 +1689,6 @@ void PackW32Pe::pack(OutputFile *fo)
     if (IDSIZE(PEDIR_COMRT))
         throwCantPack(".NET files (win32/net) are not yet supported");
 
-#if 0
-    // Structured Exception Handling
-    if (!opt->win32_pe.strip_loadconf && IDSIZE(PEDIR_LOADCONF))
-        throwCantPack("Structured Exception Handling present (try --strip-loadconf)");
-#endif
-
     if (isdll)
         opt->win32_pe.strip_relocs = false;
     else if (opt->win32_pe.strip_relocs < 0)
@@ -1751,12 +1774,14 @@ void PackW32Pe::pack(OutputFile *fo)
 
     Resource res;
     Interval tlsiv(ibuf);
+    Interval loadconfiv(ibuf);
     Export xport((char*)(unsigned char*)ibuf);
 
     const unsigned dllstrings = processImports();
     processTls(&tlsiv); // call before processRelocs!!
     processResources(&res);
     processExports(&xport);
+    processLoadConf(&loadconfiv);
     processRelocs();
 
     //OutputFile::dump("x1", ibuf, usize);
@@ -1879,6 +1904,45 @@ void PackW32Pe::pack(OutputFile *fo)
         unsigned jmp_pos = find_le32(loader,codesize + 4,get_le32("JMPO"));
         patch_le32(loader,codesize + 4,"JMPO",ih.entry - upxsection - jmp_pos - 4);
     }
+    if (use_dep_hack)
+    {
+        // This works around a "protection" introduced in MSVCRT80, which
+        // works like this:
+        // When the compiler detects that it would link in some code from its
+        // C runtime library which references some data in a read only
+        // section then it compiles in a runtime check whether that data is
+        // still in a read only section by looking at the pe header of the
+        // file. If this check fails the runtime does "interesting" things
+        // like not running the floating point initialization code - the result
+        // is an R6002 runtime error.
+        // These supposed to be read only addresses are covered by the sections
+        // UPX0 & UPX1 in the compressed files, so we have to patch the PE header
+        // in the memory. And the page on which the PE header is stored is read
+        // only so we must make it rw, fix the flags (i.e. clear
+        // PEFL_WRITE of osection[x].flags), and make it ro again.
+
+        // rva of the most significant byte of member "flags" in section "UPX0"
+        const unsigned swri = pe_offset + sizeof(oh) + sizeof(pe_section_t) - 1;
+#if 0
+        if (swri >= 0x1000)
+            throwCantPack("swri >= 0x1000! Send a bug report please!");
+        patch_le32(loader, codesize, "SWRI", swri);
+        patch_le32(loader, codesize, "IMGL", 0x1000);           // one page
+        patch_le32(loader, codesize, "IMGB", 0u - rvamin);
+#else
+        // make sure we only touch the minimum number of pages
+        const unsigned addr = 0u - rvamin + swri;
+        patch_le32(loader, codesize, "SWRI", addr &  0xfff);    // page offset
+        // check whether osection[0].flags and osection[1].flags
+        // are on the same page
+        if ((addr & 0xfff) + 0x28 >= 0x1000)
+            patch_le32(loader, codesize, "IMGL", 0x2000);       // two pages
+        else
+            patch_le32(loader, codesize, "IMGL", 0x1000);       // one page
+        patch_le32(loader, codesize, "IMGB", addr &~ 0xfff);    // page mask
+#endif
+        patch_le32(loader, codesize, "VPRO", myimport + get_le32(oimpdlls + 16) + 8);
+    }
     if (big_relocs & 6)
         patch_le32(loader,codesize,"DELT", 0u - (unsigned) ih.imagebase - rvamin);
     if (sorelocs && (soimport == 0 || soimport + cimports != crelocs))
@@ -1886,7 +1950,7 @@ void PackW32Pe::pack(OutputFile *fo)
     if (soimport)
     {
         if (!isdll)
-            patch_le32(loader,codesize,"EXIT",myimport + get_le32(oimpdlls + 16) + 8);
+            patch_le32(loader,codesize,"EXIT",myimport + get_le32(oimpdlls + 16) + 12);
         patch_le32(loader,codesize,"GETP",myimport + get_le32(oimpdlls + 16) + 4);
         if (kernel32ordinal)
             patch_le32(loader,codesize,"K32O",myimport);
@@ -1944,14 +2008,6 @@ void PackW32Pe::pack(OutputFile *fo)
     ODADDR(PEDIR_BOUNDIM) = 0;
     ODSIZE(PEDIR_BOUNDIM) = 0;
 
-#if 0
-    if (opt->win32_pe.strip_loadconf)
-    {
-        ODADDR(PEDIR_LOADCONF) = 0;
-        ODSIZE(PEDIR_LOADCONF) = 0;
-    }
-#endif
-
     // tls is put into section 1
 
     ic = s1addr + s1size - sotls;
@@ -1984,7 +2040,7 @@ void PackW32Pe::pack(OutputFile *fo)
     }
     ic += soexport;
 
-    processLoadConf();
+    processLoadConf(&rel, &loadconfiv, ic);
     ODADDR(PEDIR_LOADCONF) = soloadconf ? ic : 0;
     ODSIZE(PEDIR_LOADCONF) = soloadconf;
     ic += soloadconf;
@@ -2070,6 +2126,8 @@ void PackW32Pe::pack(OutputFile *fo)
     fo->write(obuf,clen);
     infoWriting("compressed data", clen);
     fo->write(loader,codesize);
+    if (opt->debug.dump_stub_loader)
+        OutputFile::dump(opt->debug.dump_stub_loader, loader, codesize);
     if ((ic = fo->getBytesWritten() & 3) != 0)
         fo->write(ibuf,4 - ic);
     fo->write(otls,sotls);
