@@ -2,8 +2,8 @@
 
    This file is part of the UPX executable compressor.
 
-   Copyright (C) 1996-2005 Markus Franz Xaver Johannes Oberhumer
-   Copyright (C) 1996-2005 Laszlo Molnar
+   Copyright (C) 1996-2006 Markus Franz Xaver Johannes Oberhumer
+   Copyright (C) 1996-2006 Laszlo Molnar
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -295,7 +295,7 @@ bool Packer::checkFinalCompressionRatio(const OutputFile *fo) const
 **************************************************************************/
 
 void Packer::decompress(const upx_bytep in, upx_bytep out,
-                        bool verify_checksum)
+                        bool verify_checksum, Filter *ft)
 {
     unsigned adler;
 
@@ -316,6 +316,9 @@ void Packer::decompress(const upx_bytep in, upx_bytep out,
     // verify checksum of decompressed data
     if (verify_checksum)
     {
+        if (ft) {
+            ft->unfilter(out, ph.u_len);
+        }
         adler = upx_adler32(out, ph.u_len, ph.saved_u_adler);
         if (adler != ph.u_adler)
             throwChecksumError();
@@ -339,9 +342,9 @@ bool Packer::testOverlappingDecompression(const upx_bytep buf,
         return false;
 
     assert((int)overlap_overhead >= 0);
-    // Because we are not using the asm_fast decompressor here
-    // we must account for extra 3 bytes or else we may fail
-    // at runtime decompression.
+    // Because upx_test_overlap() does not use the asm_fast decompressor
+    // we must account for extra 3 bytes that asm_fast does use,
+    // or else we may fail at runtime decompression.
     if (overlap_overhead <= 4 + 3)  // don't waste time here
         return false;
     overlap_overhead -= 3;
@@ -355,7 +358,7 @@ bool Packer::testOverlappingDecompression(const upx_bytep buf,
 }
 
 
-void Packer::verifyOverlappingDecompression()
+void Packer::verifyOverlappingDecompression(Filter *ft)
 {
     assert(ph.c_len < ph.u_len);
     assert((int)ph.overlap_overhead > 0);
@@ -381,7 +384,7 @@ void Packer::verifyOverlappingDecompression()
     if (offset + ph.c_len > obuf.getSize())
         return;
     memmove(obuf + offset, obuf, ph.c_len);
-    decompress(obuf + offset, obuf, true);
+    decompress(obuf + offset, obuf, true, ft);
     obuf.checkState();
 #endif /* !UNUPX */
 }
@@ -812,7 +815,26 @@ int Packer::patchVersion(void *b, int blen)
     checkPatch(b, blen, boff, 14);
 
     unsigned char *p = (unsigned char *)b + boff + 9;
-    memcpy(p, UPX_VERSION_STRING4, 4);
+    if (opt->debug.fake_stub_version[0])
+        memcpy(p, opt->debug.fake_stub_version, 4);
+    else
+        memcpy(p, UPX_VERSION_STRING4, 4);
+
+    return boff;
+}
+
+
+// patch year into stub/ident_[ns].ash
+int Packer::patchVersionYear(void *b, int blen)
+{
+    int boff = find(b, blen, " 1996-UPXY ", 11);
+    checkPatch(b, blen, boff, 11);
+
+    unsigned char *p = (unsigned char *)b + boff + 6;
+    if (opt->debug.fake_stub_year[0])
+        memcpy(p, opt->debug.fake_stub_year, 4);
+    else
+        memcpy(p, UPX_VERSION_YEAR, 4);
 
     return boff;
 }
@@ -976,21 +998,43 @@ const int *Packer::getDefaultCompressionMethods_le32(int method, int level, int
 
 char const *Packer::getIdentstr(unsigned *size, int small)
 {
-    static const char identbig[] =
+    static char identbig[] =
         "\n\0"
         "$Info: "
         "This file is packed with the UPX executable packer http://upx.sf.net $"
         "\n\0"
         "$Id: UPX "
         UPX_VERSION_STRING4
-        " Copyright (C) 1996-2005 the UPX Team. All Rights Reserved. $"
+        " Copyright (C) 1996-" UPX_VERSION_YEAR " the UPX Team. All Rights Reserved. $"
         "\n";
-    static const char identsmall[] =
+    static char identsmall[] =
         "\n"
         "$Id: UPX "
-        "(C) 1996-2005 the UPX Team. All Rights Reserved. http://upx.sf.net $"
+        "(C) 1996-" UPX_VERSION_YEAR " the UPX Team. All Rights Reserved. http://upx.sf.net $"
         "\n";
-    static const char identtiny[] = UPX_VERSION_STRING4;
+    static char identtiny[] = UPX_VERSION_STRING4;
+
+    static int done;
+    if (!done && (opt->debug.fake_stub_version[0] || opt->debug.fake_stub_year[0]))
+    {
+        struct strinfo_t { char *s; int size; };
+        static const strinfo_t strlist[] = {
+            { identbig,   (int)sizeof(identbig) },
+            { identsmall, (int)sizeof(identsmall) },
+            { identtiny,  (int)sizeof(identtiny) },
+        { NULL, 0 } };
+        const strinfo_t* iter;
+
+        for (iter = strlist; iter->s; ++iter)
+        {
+            if (opt->debug.fake_stub_version[0])
+                mem_replace(iter->s, iter->size, UPX_VERSION_STRING4, 4, opt->debug.fake_stub_version);
+            if (opt->debug.fake_stub_year[0])
+                mem_replace(iter->s, iter->size, UPX_VERSION_YEAR, 4, opt->debug.fake_stub_year);
+        }
+        done = 1;
+    }
+
 
     if (small < 0)
         small = opt->small;
@@ -1033,13 +1077,15 @@ void Packer::initLoader(const void *pdata, int plen, int pinfo, int small)
 
 void __acc_cdecl_va Packer::addLoader(const char *s, ...)
 {
-    const char *p;
     va_list ap;
 
     va_start(ap, s);
-    linker->addSection(s);
-    while((p = va_arg(ap, const char *)) != NULL)
-        linker->addSection(p);
+    while (s != NULL)
+    {
+        if (*s)
+            linker->addSection(s);
+        s = va_arg(ap, const char *);
+    }
     va_end(ap);
 }
 
@@ -1215,13 +1261,32 @@ const char *Packer::getDecompressor() const
 // executable formats.
 //
 // It will replace the tryFilters() / compress() call sequence.
+
+// 2006-02-15: hdr_buf and hdr_u_len are default empty input "header" array
+// to fix a 2-pass problem with Elf headers.  As of today there can be
+// only one decompression method per executable output file, and that method
+// is the one that gives best compression for .text and loader.  However,
+// the Elf headers precede .text in the output file, and are written first.
+// "--brute" compression often compressed the Elf headers using nrv2b
+// but the .text (and loader) with nrv2e.  This often resulted in SIGSEGV
+// during decompression.
+// The workaround is for hdr_buf and hdr_u_len to describe the Elf headers
+// (typically less than 512 bytes) when .text is passed in, and include
+// them in the calculation of shortest output.  Then the result
+// this->ph.method  will say which [single] method to use for everthing.
+// The Elf headers are never filtered.  They are short enough (< 512 bytes)
+// that compressing them more than once per method (once here when choosing,
+// once again just before writing [because compressWithFilters discards])
+// is OK because of the simplicity of not having two output arrays.
 **************************************************************************/
 
 void Packer::compressWithFilters(Filter *parm_ft,
                                  const unsigned overlap_range,
                                  int strategy, const int *parm_filters,
                                  unsigned max_offset, unsigned max_match,
-                                 unsigned filter_off, unsigned compress_buf_off)
+                                 unsigned filter_off, unsigned compress_buf_off,
+                                 unsigned char *hdr_buf,
+                                 unsigned hdr_u_len)
 {
     const int *f;
     //
@@ -1237,6 +1302,7 @@ void Packer::compressWithFilters(Filter *parm_ft,
     best_ph.c_len = orig_ph.u_len;
     best_ph.overlap_overhead = 0;
     unsigned best_ph_lsize = 0;
+    unsigned best_hdr_clen = 0;
 
     // preconditions
     assert(orig_ph.filter == 0);
@@ -1330,6 +1396,19 @@ void Packer::compressWithFilters(Filter *parm_ft,
     int nfilters_success = 0;
     for (int m = 0; m < nmethods; m++)          // for all methods
     {
+        unsigned hdr_clen = 0;
+        if (hdr_buf && hdr_u_len) {
+            unsigned result[16];
+            upx_compress_config_t conf;
+            memset(&conf, 0xff, sizeof(conf));
+            if (0 < m && otemp == &obuf) { // do not overwrite obuf
+                otemp_buf.allocForCompression(compress_buf_len);
+                otemp = &otemp_buf;
+            }
+            int r = upx_compress(hdr_buf, hdr_u_len, *otemp, &hdr_clen,
+                0, methods[m], 10, &conf, result);
+            (void)r;
+        }
         for (int i = 0; i < nfilters; i++)          // for all filters
         {
             ibuf.checkState();
@@ -1385,19 +1464,19 @@ void Packer::compressWithFilters(Filter *parm_ft,
                     lsize = buildLoader(&ft);
                 }
 #if 0
-                printf("\n%2d %02x: %d +%4d = %d  (best: %d +%4d = %d)\n", ph.method, ph.filter,
-                       ph.c_len, lsize, ph.c_len + lsize,
-                       best_ph.c_len, best_ph_lsize, best_ph.c_len + best_ph_lsize);
+                printf("\n%2d %02x: %d +%4d +%3d = %d  (best: %d +%4d +%3d = %d)\n", ph.method, ph.filter,
+                       ph.c_len, lsize, hdr_clen, ph.c_len + lsize + hdr_clen,
+                       best_ph.c_len, best_ph_lsize, best_hdr_clen, best_ph.c_len + best_ph_lsize + best_hdr_clen);
 #endif
                 bool update = false;
-                if (ph.c_len + lsize < best_ph.c_len + best_ph_lsize)
+                if (ph.c_len + lsize + hdr_clen < best_ph.c_len + best_ph_lsize + best_hdr_clen)
                     update = true;
-                else if (ph.c_len + lsize == best_ph.c_len + best_ph_lsize)
+                else if (ph.c_len + lsize + hdr_clen == best_ph.c_len + best_ph_lsize + best_hdr_clen)
                 {
                     // prefer smaller loaders
-                    if (lsize < best_ph_lsize)
+                    if (lsize  + hdr_clen < best_ph_lsize + best_hdr_clen)
                         update = true;
-                    else if (lsize == best_ph_lsize)
+                    else if (lsize + hdr_clen == best_ph_lsize + best_hdr_clen)
                     {
                         // prefer less overlap_overhead
                         if (ph.overlap_overhead < best_ph.overlap_overhead)
@@ -1412,6 +1491,7 @@ void Packer::compressWithFilters(Filter *parm_ft,
                     // save compression results
                     best_ph = ph;
                     best_ph_lsize = lsize;
+                    best_hdr_clen = hdr_clen;
                     best_ft = ft;
                 }
             }
